@@ -1,9 +1,9 @@
 #' @param attributes If TRUE (default), attributes of `expr` are also searched.
 #' If FALSE, they are not.
 #' If a character vector, then attributes with matching names are searched.
-#' Note, the attributes of the attributes elements are not searched, that is,
+#' Note, the attributes of the attributes' elements are not searched, that is,
 #' attributes are not searched recursively.  Also, attributes are searched
-#' with `dotdotdot = "ignore".
+#' with `dotdotdot = "ignore"`.
 #'
 #' @param dotdotdot TBD.
 #'
@@ -17,19 +17,63 @@ findGlobals <- function(expr, envir = parent.frame(), ...,
                         attributes = TRUE,
                         tweak = NULL,
                         dotdotdot = c("warning", "error", "return", "ignore"),
-                        method = c("ordered", "conservative", "liberal"),
+                        method = c("ordered", "conservative", "liberal", "dfs"),
                         substitute = FALSE, unlist = TRUE, trace = FALSE) {
-  method <- match.arg(method, choices = c("ordered", "conservative", "liberal"))
+  if (missing(method)) method <- method[1]                        
+  method <- match.arg(method, choices = c("ordered", "conservative", "liberal", "dfs"), several.ok = TRUE)
   dotdotdot <- match.arg(dotdotdot, choices = c("warning", "error", "return", "ignore"))
 
   if (substitute) expr <- substitute(expr)
 
   if (trace) {
-    trace_msg <- trace_enter("findGlobals(..., dotdotdot = '%s', method = '%s', unlist = %s)", dotdotdot, method, unlist)
+    methods <- sprintf("'%s'", method)
+    if (length(method) > 1) methods <- sprintf("c(%s)", paste(methods, collapse = ", "))
+    trace_msg <- trace_enter("findGlobals(..., dotdotdot = '%s', method = %s, unlist = %s)", dotdotdot, methods, unlist)
     on.exit(trace_exit(trace_msg))
   }
 
-  debug <- mdebug("findGlobals(..., dotdotdot = '%s', method = '%s', unlist = %s) ...", dotdotdot, method, unlist)
+  debug <- isTRUE(getOption("globals.debug"))
+  if (debug) {
+    methods <- sprintf("'%s'", method)
+    if (length(method) > 1) methods <- sprintf("c(%s)", paste(methods, collapse = ", "))
+    mdebugf_push("findGlobals(..., dotdotdot = '%s', method = %s, unlist = %s) ...", dotdotdot, methods, unlist)
+    on.exit(mdebugf_pop("findGlobals(..., dotdotdot = '%s', method = %s, unlist = %s) ... done", dotdotdot, methods, unlist), add = TRUE)
+  }
+
+  if (length(method) > 1) {
+    if (!unlist) {
+      stop("Argument 'unlist' must be TRUE if more than one 'method' is specified: ", commaq(method))
+    }
+    if (is.function(tweak)) {
+      if (debug) mdebug("tweaking expression using function")
+      expr <- tweak(expr)
+    }
+    globals <- list()
+    for (mtd in method) {
+      globals[[mtd]] <- findGlobals(
+        expr, substitute = FALSE, envir = envir, ...,
+        attributes = attributes, tweak = NULL,
+        dotdotdot = dotdotdot, method = mtd,
+        unlist = TRUE, trace = trace
+      )
+    }
+
+    globals_all <- unlist(globals, use.names = FALSE)
+    globals_all <- globals_all[!duplicated(globals_all)]
+
+    if (debug) {
+      mdebug("Globals found by per method:")
+      w <- max(nchar(method)) + 2L
+      for (mtd in method) {
+        names <- globals[[mtd]]
+        delta <- setdiff(globals_all, names)
+        delta <- if (length(delta) == 0) "<none>" else commaq(delta)
+        mdebugf("%*s: %s [delta: %s]", w, sQuote(mtd), commaq(names), delta)
+      }
+    }
+
+    return(globals_all)
+  } ## if (length(method) > 1)
 
   if (is.logical(attributes)) {
     stop_if_not(length(attributes) == 1L, !is.na(attributes))
@@ -39,10 +83,10 @@ findGlobals <- function(expr, envir = parent.frame(), ...,
   }
   
   if (is.list(expr)) {
-    debug && mdebug(" - expr: <a list of length %d>", .length(expr))
+    if (debug) mdebugf("expr: <a list of length %d>", .length(expr))
 
     ## NOTE: Do *not* look for types that we are interested in, but instead
-    ## look for types that we are *not* interested.  The reason for this that
+    ## look for types that we are *not* interested in.  The reason for this is that
     ## in future versions of R there might be new types added that may contain
     ## globals and with this approach those types will also be scanned.
     basicTypes <- c("logical", "integer", "double", "complex", "character",
@@ -50,20 +94,15 @@ findGlobals <- function(expr, envir = parent.frame(), ...,
 
     ## Skip elements in 'expr' of basic types that cannot contain globals
     types <- unlist(list_apply(expr, FUN = typeof), use.names = FALSE)
-    keep <- !(types %in% basicTypes)
-
-    ## Don't use expr[keep] here, because that may use S3 dispatching
-    ## depending on class(expr)
-    expr <- .subset(expr, keep)
+    keep <- which(!(types %in% basicTypes))
 
     ## Early stopping?
-    if (.length(expr) == 0) {
-      debug && mdebug(" - globals found: [0] <none>")
-      debug && mdebug("findGlobals(..., dotdotdot = '%s', method = '%s', unlist = %s) ... DONE", dotdotdot, method, unlist) #nolint
+    if (length(keep) == 0) {
+      if (debug) mdebug("globals found: [0] <none>")
       return(character(0L))
     }
-    
-    globals <- list_apply(expr, FUN = findGlobals, envir = envir,
+
+    globals <- list_apply(expr, subset = keep, FUN = findGlobals, envir = envir,
                       attributes = attributes, ...,
                       tweak = tweak, dotdotdot = dotdotdot,
                       method = method,
@@ -72,7 +111,7 @@ findGlobals <- function(expr, envir = parent.frame(), ...,
     
     keep <- types <- NULL ## Not needed anymore
     
-    debug && mdebug(" - preliminary globals found: [%d] %s",
+    if (debug) mdebugf("preliminary globals found: [%d] %s",
                     length(globals), hpaste(sQuote(names(globals))))
 
     if (unlist) {
@@ -83,28 +122,47 @@ findGlobals <- function(expr, envir = parent.frame(), ...,
       if (length(idxs) > 0L) globals <- c(globals[-idxs], globals[idxs])
     }
 
-    debug && mdebug(" - globals found: [%d] %s",
+    if (debug) mdebugf("globals found: [%d] %s",
                     length(globals), hpaste(sQuote(globals)))
-    debug && mdebug("findGlobals(..., dotdotdot = '%s', method = '%s', unlist = %s) ... DONE", dotdotdot, method, unlist) #nolint
-    
     return(globals)
   }
 
   if (is.function(tweak)) {
-    debug && mdebug(" - tweaking expression using function")
+    if (debug) mdebug("tweaking expression using function")
     expr <- tweak(expr)
   }
 
-  if (method == "ordered") {
-    find_globals_t <- find_globals_ordered
-  } else if (method == "conservative") {
-    find_globals_t <- find_globals_conservative
-  } else if (method == "liberal") {
-    find_globals_t <- find_globals_liberal
-  }
+  if (method == "dfs") {
+    globals <- findGlobalsDFS(expr)
+  } else {
+    if (hasCodetoolsBug16()) {
+      if (debug) mdebug("workaround 'codetools' bug #16")
+      expr <- walkAST(expr, call = tweakCodetoolsBug16)
+    }
 
-  globals <- call_find_globals_with_dotdotdot(find_globals_t, expr = expr, envir = envir, dotdotdot = dotdotdot, trace = trace, debug = debug)
+    if (is.expression(expr)) {
+      return(findGlobals(expr[[1]], substitute = substitute, envir = envir,
+                         attributes = attributes, tweak = tweak, ...,
+                         dotdotdot = dotdotdot, method = method,
+                         unlist = unlist, trace = trace))
+    }
+
+    if (method == "ordered") {
+      find_globals_t <- find_globals_ordered
+    } else if (method == "conservative") {
+      find_globals_t <- find_globals_conservative
+    } else if (method == "liberal") {
+      find_globals_t <- find_globals_liberal
+    }
   
+    globals <- call_find_globals_with_dotdotdot(find_globals_t, expr = expr, envir = envir, dotdotdot = dotdotdot, trace = trace, debug = debug)
+    idx <- which(globals == "codetools.bugfix16:::$<-")
+    if (length(idx) > 0) {
+      globals[idx] <- "$<-"
+      globals <- unique(globals)
+    }
+  } ## if (method == ...)
+
   ## Search attributes?
   if (length(attributes) > 0) {
     attrs <- attributes(expr)
@@ -114,7 +172,7 @@ findGlobals <- function(expr, envir = parent.frame(), ...,
 
     ## Attributes to be searched, if any
     if (length(attrs) > 0) {
-      debug && mdebug(" - searching attributes")
+      if (debug) mdebug("searching attributes")
       attrs_globals <- list_apply(attrs, FUN = findGlobals, envir = envir,
                                   ## Don't look for attributes recursively
                                   attributes = FALSE,
@@ -127,14 +185,13 @@ findGlobals <- function(expr, envir = parent.frame(), ...,
                                   trace = trace)
       if (unlist) attrs_globals <- unlist(attrs_globals, use.names = FALSE)
       if (length(attrs_globals) > 1L) attrs_globals <- unique(attrs_globals)
-      debug && mdebug(" - globals found in attributes: [%d] %s",
+      if (debug) mdebugf("globals found in attributes: [%d] %s",
                       length(attrs_globals), hpaste(sQuote(attrs_globals)))
       globals <- unique(c(globals, attrs_globals))
     }
   }
-
-  debug && mdebug(" - globals found: [%d] %s", length(globals), hpaste(sQuote(globals)))
-  debug && mdebug("findGlobals(..., dotdotdot = '%s', method = '%s', unlist = %s) ... DONE", dotdotdot, method, unlist) #nolint
+  
+  if (debug) mdebugf("globals found: [%d] %s", length(globals), hpaste(sQuote(globals)))
 
   globals
 }
